import os
import sys
import signal
import logging
import argparse
import functools
import subprocess
from datetime import timedelta
from gpstime import gpstime, GPSTimeException
from dateutil.tz import tzutc, tzlocal

from . import systemd
from . import svlogd
from . import util

##################################################

PROG = 'guardctrl'

try:
    IFO = os.environ['IFO']
except KeyError:
    sys.exit("Must specify IFO.")
try:
    SITE = os.environ['SITE']
except KeyError:
    sys.exit("Must specify SITE.")

##################################################

parser = argparse.ArgumentParser(
    prog = 'guardctrl',
    description = """Guardian daemon supervision control interface.

Control guardian daemon processes managed by the site systemd
supervision system.
""",
    formatter_class = argparse.RawDescriptionHelpFormatter,
    epilog="""
Add '-h' after individual commands for command help.

Node names may be specified with wildcard/globbing, e.g. 'SUS_*'.
A single '*' will act on all configured nodes (where appropriate).
""",
)

parser.add_argument('-d', '--debug', action='store_true',
                    help="print debug information to stderr")
# parser.add_argument('-v', '--version', action='version',
#                     version='Guardctrl {}'.format(__version__),
#                     help="print guardctrl version and exit")

subparser = parser.add_subparsers(
    title='Commands',
    metavar='<command>',
    dest='cmd',
    #help=argparse.SUPPRESS,
)
subparser.required = True

def gen_subparser(cmd, func, node_nargs=None):
    help = func.__doc__.split('\n')[0].lower().strip('.')
    desc = func.__doc__
    p = subparser.add_parser(
        cmd,
        #formatter_class=argparse.RawDescriptionHelpFormatter,
        help=help,
        description=desc,
    )
    p.set_defaults(func=func)
    if node_nargs:
        p.add_argument('nodes', metavar='node', nargs=node_nargs,
                       help="guardian node name (may be wildcard/glob pattern)")
    return p

##################################################

def command_help(args):
    """Show command help."""
    if args.cmd:
        try:
            p = subparser.choices[args.cmd]
        except KeyError:
            sys.exit("Unknown command: {}".format(args.cmd))
    else:
        p = parser
    p.print_help()

p = gen_subparser("help", command_help)
p.add_argument('cmd', metavar='<command>', nargs='?', help="command")


def list_nodes(args):
    """List nodes and node subervision state."""
    nodes = args.nodes
    include_status = args.include_status
    only_active = args.only_active
    if not nodes:
        nodes = '*'
    nodes = systemd.list_nodes(nodes,
                               only_active=only_active,
                               include_status=True)
    if not nodes:
        return
    node_w = len(max([node[0] for node in nodes])) + 4
    if include_status:
        fmt = '{{node:<{node_w}}} {{enabled:<{enabled_w}}} {{active:<{active_w}}}'.format(
            node_w=node_w,
            enabled_w=10,
            active_w=10,
        )
    else:
        fmt = '{node}'
    for node, enabled, active in nodes:
        kwargs = {}
        if active == 'failed':
            kwargs = dict(fg='red', bold=True)
        elif active != 'active':
            kwargs = dict(fg='yellow')
        elif active == 'active':
            kwargs = dict(fg='green')
        line = fmt.format(node=node, enabled=enabled, active=active)
        print(line, **kwargs)

p = gen_subparser('list', list_nodes, node_nargs='*')
p.add_argument('-s', '--status', dest='include_status', action='store_true',
               help="")
p.add_argument('-r', '--running', dest='only_active', action='store_true',
               help="")


@functools.wraps(systemd.print_node_status)
def print_node_status(args):
    nodes = args.nodes
    try:
        return systemd.print_node_status(nodes, print_logs=True)
    except util.GuardCtrlError as e:
        sys.exit(e)

gen_subparser('status', print_node_status, node_nargs='*')


@functools.wraps(systemd.enable_nodes)
def enable_nodes(args):
    nodes = args.nodes
    try:
        return systemd.enable_nodes(nodes)
    except util.GuardCtrlError as e:
        sys.exit(e)

gen_subparser('enable', enable_nodes, node_nargs='+')


@functools.wraps(systemd.start_nodes)
def start_nodes(args):
    nodes = args.nodes
    try:
        return systemd.start_nodes(nodes)
    except util.GuardCtrlError as e:
        sys.exit(e)

gen_subparser('start', start_nodes, node_nargs='+')


@functools.wraps(systemd.restart_nodes)
def restart_nodes(args):
    nodes = args.nodes
    try:
        return systemd.restart_nodes(nodes)
    except util.GuardCtrlError as e:
        sys.exit(e)

gen_subparser('restart', restart_nodes, node_nargs='+')


@functools.wraps(systemd.stop_nodes)
def stop_nodes(args):
    nodes = args.nodes
    try:
        return systemd.stop_nodes(nodes)
    except util.GuardCtrlError as e:
        sys.exit(e)

gen_subparser('stop', stop_nodes, node_nargs='+')


@functools.wraps(systemd.disable_nodes)
def disable_nodes(args):
    nodes = args.nodes
    node_list = systemd.list_nodes(nodes)
    if not node_list:
        sys.exit("Unknown nodes: {}".format(' '.join(nodes)))
    print("Really disable the following nodes?", file=sys.stderr)
    for node in node_list:
        print("  {}".format(node), file=sys.stderr)
    print("This will remove these nodes from the site list.", file=sys.stderr)
    ret = input("Type 'yes' to disable: ")
    if ret == 'yes':
        return
    try:
        return systemd.disable_nodes(node_list)
    except util.GuardCtrlError as e:
        sys.exit(e)

gen_subparser('disable', disable_nodes, node_nargs='+')

##################################################

def print_logs(args):
    """View node logs.

Log timestamps are in UTC by default."""
    node = args.nodes
    time_format = args.time_format
    search_old_logs = args.search_old_logs
    # ddelta = timedelta(seconds=float(args.duration))

    # if carg('hours'):
    #     args.after = gpstime.now(tz=tzlocal()) - timedelta(seconds=3600*float(args.hours))
    # elif carg('after'):
    #     args.after = _parse_time_arg('after')
    #     if carg('before'):
    #         args.before = _parse_time_arg('before')
    #         if args.before < args.after:
    #             raise util.GuardCtrlError("Before time is earlier than after time.")
    #     elif carg('duration'):
    #         args.before = args.after + ddelta
    # elif carg('center'):
    #     dur = ddelta/2
    #     args.after = _parse_time_arg('center') - dur
    #     args.before = _parse_time_arg('center') + dur
    # elif carg('before'):
    #     args.before = _parse_time_arg('before')
    #     args.after = args.before - ddelta
    # # else:
    # #     args.follow = True

    # logging.debug("search after : %s" % args.after)
    # logging.debug("search before: %s" % args.before)

    if time_format == 'utc':
        def dt2s(dt):
            return dt.astimezone(tzutc()).strftime(util.ISO_FMT+'Z')
    elif time_format == 'local':
        def dt2s(dt):
            return dt.strftime(util.ISO_FMT)
    elif time_format == 'gps':
        def dt2s(dt):
            return '{:f}'.format(gpstime.fromdatetime(dt).gps())

    if old:
        kwargs = {k: getattr(args, k) for k in ['nodes', 'after', 'before']}
        gen = svlogd.node_log_search(**kwargs)
    else:
        kwargs = {k: getattr(args, k) for k in ['nodes', 'after', 'before', 'nlines', 'follow']}
        gen = systemd.node_logs(**kwargs)

    for dt, msg in gen:
        line = '{} {}'.format(dt2s(dt), msg)
        print(line)

p = gen_subparser('log', print_logs, node_nargs='*')
p.epilog = """
Date/time strings for filters are handled by a date/time string parser
and therefore may be specified in various formats.  If --after and
--before are not both specified, --duration will be used.  --lines
will ultimately limit the number of lines returned.

Examples:

  Show all ISC_LOCK logs from Sept. 14, 2015:

  $ {prog} --after 2015-09-14 --before 2015-09-15 ISC_LOCK

  Show the last 1000 lines of all SUS node logs:

  $ {prog} -n1000 SUS_*

  Quote compound date strings, and specify 'UTC' for times in UTC:

  $ {prog} -a "2016-08-20 06:23:22 UTC" -b "2016-08-20 06:23:52 UTC" ISC_LOCK

  GPS times are accepted as well:

  $ {prog} 1155709419 ISC_DRMI

""".format(prog=PROG+' log')
p.formatter_class = argparse.RawDescriptionHelpFormatter
p.add_argument('-f', '--follow', action='store_true',
               help="follow logs")
p.add_argument('-a', '--after', metavar='DATETIME', #callback=validate_date_arg,
               help="filter for logs after specified date/time")
p.add_argument('-b', '--before', metavar='DATETIME', #callback=validate_date_arg,
               help="filter for logs after specified date/time")
#
# p.add_argument('-c', '--center', metavar='DATETIME', callback=validate_date,
#                help="filter for logs centered around specified time")
# p.add_argument('-o', '--hours', 'hours', type=int, metavar='HOURS',
#                help="filter for logs from the last H hours")
# p.add_argument('-t', '--day', 'hours', flag_value=24,
#                help="filter for logs from the last 24 hours")
# p.add_argument('-d', '--duration', metavar='SECONDS', type=float, default=60)
#
p.add_argument('-n', '--lines', metavar='N', dest='nlines', type=int, default=100,
               help="limit logs to N lines [100]")
p.add_argument('-u', '--utc', dest='time_format', action='store_const', const='utc',
               help="Show log lines with UTC timestamps (default).")
p.add_argument('-l', '--local', dest='time_format', action='store_const', const='local',
               help="Show log lines with local timestamps.")
p.add_argument('-g', '--gps', dest='time_format', action='store_const', const='gps',
               help="Show log lines with GPS timestamps.")
p.add_argument('--old', dest='search_old_logs', action='store_true',
               help="Search old logs (<=O2, some arguments ignored).")

##################################################

def main():
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)

    FORMAT = '%(levelname)s: %(module)s: %(message)s'
    logging.basicConfig(format=FORMAT, level=os.getenv('LOG_LEVEL', 'INFO').upper())

    # handle invocation as ssh ForceCommand
    if os.getenv('SSH_ORIGINAL_COMMAND') and len(sys.argv) == 1:
        sys.argv += os.getenv('SSH_ORIGINAL_COMMAND').split()

    logging.debug(sys.argv)

    # handle direct systemctl/journalctl invocation
    if len(sys.argv) > 1 and \
       sys.argv[1] in ['env', 'systemctl', 'journalctl']:
        logging.debug("cmd: {}".format(' '.join(sys.argv[1:])))
        try:
            sys.exit(subprocess.run(sys.argv[1:]).returncode)
        except KeyboardInterrupt:
            sys.exit()

    args = parser.parse_args()
    args.func(args)

if __name__ == '__main__':
    main()
